{% extends "!layout.html" %}

{% block extrahead %}
{{ super() }}
<script type='text/javascript' src='_static/x3dom.js'> </script> 
<link rel='stylesheet' type='text/css' href='_static/x3dom.css'></link> 
{% endblock %}
