# Autogenerated by Astropy-affiliated package marxs's setup.py on 2020-11-06 18:53:41 UTC
from __future__ import unicode_literals
import datetime

version = "1.2"
githash = "47f3960729640c8609aee1e4c80b552cf598248d"


major = 1
minor = 2
bugfix = 0

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2020, 11, 6, 18, 53, 41)
debug = False

astropy_helpers_version = "1.2.dev1156"
